# -*- coding: utf-8 -*-
# SDK info
__version__ = '0.1.4'
__package_name__ = "fastflyer"
__title__ = 'FastFlyer — 轻量级API开发框架'

__author__ = 'Jagerzhang'
__author_email__ = 'im@zhang.ge'

__license__ = 'Apache 2.0'

__python_requires__ = '>=3, <4'
__classifiers__ = ["Programming Language :: Python :: 3"]
__url__ = 'https://github.com/jagerzhang/fastflyer.git'

__copyright__ = 'Copyright © 2024 Tencent. All Rights Reserved. 网络平台部 平台研发中心 版权所有'
__description__ = 'FastFlyer 是基于 FastAPI 设计的轻量级 API 开发框架。在 FastAPI 优异特性的基础上集成了一系列开箱即用的组件，\
包括Redis、MySQL、Kafka等，让研发人员只需要专注于业务逻辑的实现，真正开箱即用！'



import os
import sys
import glob
import shutil
import argparse
from datetime import datetime
import pkg_resources
from fastkit.logging import get_logger

logger = get_logger(logger_name="console", log_path="/var/log")
COLOR_RED = "\033[91m"
COLOR_GREEN = "\033[92m"
COLOR_YELLOW = "\033[93m"
COLOR_DEFAULT = "\033[39m"
COLOR_RESET = "\033[0m"


def colored_cover(message, color=COLOR_YELLOW):
    return f"{color}{message}{COLOR_RESET}"


def main():
    parser = argparse.ArgumentParser(description="FastFlyer 框架开发辅助工具")
    subparsers = parser.add_subparsers(dest="command")
    create_app_parser = subparsers.add_parser("init_app", help="初始化应用代码")
    subparsers.add_parser("show_demo", help="快速启动内置示例项目")
    create_app_parser.add_argument("-f",
                                   "--force",
                                   action="store_true",
                                   help="强制覆盖已存在文件")

    args = parser.parse_args()

    if args.command == "init_app":
        # 检查是否已经初始化过，或者使用 -f 选项来强制重新创建
        if os.path.exists(".init_lock") and not args.force:
            logger.warn("项目已经被初始化过，请勿重复初始化")
        else:
            # 获取 SDK 包内 demo 目录的路径
            demo_dir = pkg_resources.resource_filename(__name__, "template")
            # 拷贝 demo 目录下的文件到当前目录
            copy_files(demo_dir, ".", args.force)
            # 如果已经初始化过，删除 .init_lock 文件
            if os.path.exists(".init_lock"):
                os.remove(".init_lock")
            # 创建并写入 .init_lock 文件
            with open(".init_lock", "w") as lock_file:
                lock_file.write(str(datetime.now()))
            logger.info(
                f"初始化完成，你可以执行 ls 命令查看目录内容或执行{colored_cover('./dev_ctrl.sh')}快速构建开发环境"
            )

    elif args.command == "show_demo":
        logger.info("尝试读取相关环境变量...")
        for key, value in os.environ.items():
            if not key.startswith("flyer_"):
                continue
            logger.info(f"{key}={value}")

        from fastflyer.tools.template.main import main_cmd
        from fastflyer.utils import get_host_ip
        port = os.getenv("flyer_port", 8080)
        prefix = os.getenv("flyer_base_url", "/flyer")
        url = f"http://{get_host_ip()}:{port}{prefix}"
        logger.info("欢迎启动体验项目，现在可以通过浏览器访问以下项目页面：")
        logger.info(f"SwaggerUI 文档：{colored_cover(f'{url}/docs')}")
        logger.info(f"ReDoc 接口文档：{colored_cover(f'{url}/redoc')}")
        main_cmd()

    else:
        # 如果没有指定命令，则输出帮助信息
        parser.print_help()


def copy_files(source_dir, dest_dir, force=False):
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)

    copied_files = []  # 用于跟踪已经拷贝过的文件路径

    for root, dirs, files in os.walk(source_dir):
        for dir in dirs:
            if dir == "__pycache__":
                continue
            src_dir = os.path.join(root, dir)
            dest_subdir = os.path.join(dest_dir,
                                       os.path.relpath(src_dir, source_dir))

            is_copied = False
            for copied_dir in copied_files:
                if dest_subdir.startswith(copied_dir):
                    is_copied = True
                    break

            if is_copied:
                continue

            if not force and os.path.exists(dest_subdir):
                logger.warn(f"目录 {dest_subdir} 已存在. 请添加 --force 参数执行强制覆盖")
            else:
                if sys.version_info >= (3, 8):
                    shutil.copytree(src_dir, dest_subdir, dirs_exist_ok=True)
                else:
                    shutil.copytree(src_dir, dest_subdir)
                copied_files.append(dest_subdir)  # 添加已拷贝的目录路径到列表

        for file in files:
            if file.endswith((".pyc", ".log", ".lock")):
                continue  # 排除后缀为 .pyc 和 .log 的文件
            src_file = os.path.join(root, file)
            dest_file = os.path.join(dest_dir,
                                     os.path.relpath(src_file, source_dir))

            is_copied = False
            for copied_dir in copied_files:
                if dest_file.startswith(copied_dir):
                    is_copied = True
                    break

            if is_copied:
                continue

            if not force and os.path.exists(dest_file):
                logger.warn(f"文件 {dest_file} 已存在. 请添加 --force 参数执行强制覆盖")
            else:
                shutil.copy2(src_file, dest_file)

    # 删除当前目录下以 .log 结尾的文件
    for file in glob.glob("*.log"):
        os.remove(file)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import time
import json
from uuid import uuid4
import resource
from os import getenv
from fastapi.routing import APIRoute
from fastapi import Request, Response
from fastflyer.utils import get_host_ip, get_client_ip
from fastflyer import logger

host_ip = get_host_ip()


class MiddleWare(APIRoute):
    """路由中间件：记录请求日志和耗时等公共处理逻辑
    """

    async def _get_request_body(self, request: Request):
        """获取请求参数
        """
        request_body = await request.body()
        request_body = bytes.decode(request_body)

        return request_body

    async def _report_log(self, **kwargs):
        """上报智研，依赖 settings 初始化
        """
        if int(getenv("flyer_access_log", 1)) == 0:
            return

        try:
            kwargs["response"]["body"] = bytes.decode(
                kwargs["response"]["body"])

        except Exception:  # pylint: disable=broad-except
            kwargs["response"]["body"] = str(kwargs["response"]["body"])

        try:
            # 优先支持日志汇字段结构化解析
            logger.info(json.dumps(kwargs))
        except Exception:  # pylint: disable=broad-except
            logger.warning(kwargs)

    def get_route_handler(self):
        original_route_handler = super().get_route_handler()

        async def recorder(request: Request) -> Response:
            start_time = time.perf_counter()
            # 计算线程内存占用
            memory_usage_begin = resource.getrusage(
                resource.RUSAGE_THREAD).ru_maxrss

            # 对接NGate网关记录客户端ID
            client_id = request.headers.get("x-client-id", "")
            request_id = request.headers.get("x-request-id") or str(uuid4())
            client_ip = get_client_ip(request)

            # FastAPI 0.68.2 开始必须要传入 application/json 才识别为字典，否则报错，这里先兼容下，需要推动客户端整改后去掉
            # https://github.com/tiangolo/fastapi/releases/tag/0.65.2
            request.headers.__dict__["_list"].insert(
                0, (b"content-type", b"application/json"))
            request.headers.__dict__["_list"].insert(
                0, (b"x-request-id", request_id.encode()))

            response: Response = await original_route_handler(request)

            # 插入自定义头部
            memory_usage_end = resource.getrusage(
                resource.RUSAGE_THREAD).ru_maxrss
            memory_usage = memory_usage_end - memory_usage_begin
            latency = int((time.perf_counter() - start_time) * 1000)
            response.headers["X-Lasting-Time"] = str(latency)
            response.headers["X-Request-ID"] = request_id
            response.headers["X-Host-IP"] = host_ip
            response.headers["X-Memory-Usage"] = f"{memory_usage}KB"
            # 防御 XSS 反射型漏洞
            response.headers["X-Content-Type-Options"] = "nosniff"

            access_log = {
                "direction": "in",
                "request": {
                    "method": str(request.method),
                    "url": str(request.url),
                    "body": await self._get_request_body(request),
                    "headers": dict(request.headers.items())
                },
                "response": {
                    "status_code": response.status_code,
                    "body": response.body,
                    "headers": dict(response.headers.items())
                },
                "latency": latency,
                "clientIp": client_ip,
                "clientId": str(client_id),
                "memoryUsage": memory_usage,
                "logId": request_id
            }

            await self._report_log(**access_log)
            return response

        return recorder



import os
import sys
import time
import importlib
import traceback
import asyncio
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
from fastapi import FastAPI, Depends
from fastapi.staticfiles import StaticFiles
from contextlib import asynccontextmanager
from fastflyer.exceptions import init_exception
from fastflyer import config, background_scheduler, asyncio_scheduler
from fastflyer.docs import router as docs
from fastflyer.authorize import authorize
from fastkit.logging import get_logger

static_dir = os.path.join(os.path.dirname(__file__), "static")
logger = get_logger(logger_name="console", log_path=config.LOG_PATH)


class FlyerAPI:
    _inst = None

    def __new__(cls, app_path):
        """
        规避重复加载
        """
        if cls._inst is None:
            cls.app_path = app_path
            cls._inst = cls.create_app()

        return cls._inst

    @classmethod
    def create_app(cls):
        """创建应用"""
        cls.app = FastAPI(title=config.API_TITLE,
                          description=config.DESCRIPTION,
                          version=config.VERSION,
                          openapi_url=config.PREFIX + "/openapi.json",
                          docs_url=None,
                          redoc_url=None,
                          lifespan=lifespan)
        cls.app.mount(config.PREFIX + "/static",
                      StaticFiles(directory=static_dir),
                      name="static")
        cls.app.include_router(docs)
        # 自动加载子项目
        cls.load_module()
        # 初始化异常处理
        init_exception(cls.app)
        return cls.app

    @classmethod
    def load_module(cls):
        """
        加载子项目

        Args:
            app (FastAPI): FastAPI对象
        """
        # 获取指定项目的根目录，即app目录的上一层
        root_path = os.path.dirname(cls.app_path)
        # 获取app目录名称
        app_dir = os.path.basename(cls.app_path)
        # 进入项目根目录
        sys.path.append(root_path)
        os.chdir(root_path)
        # 自动加载有路由的包
        for dir in Path(app_dir).iterdir():
            # 不存在则跳过
            if not dir.exists():
                continue

            # 隐藏文件夹或申明非开放文件夹或不是文件夹的跳过
            if dir.name.startswith("_") or dir.name.startswith(
                    ".") or not dir.is_dir():
                continue

            try:
                # eval(f"import {app_dir}.{dir.name}")
                sub_module = importlib.import_module(f"{app_dir}.{dir.name}")
                # 尝试获取子模块是否启用，默认为True
                sub_module_enabled = getattr(sub_module, "__ENABLED__", True)
                if not sub_module_enabled:
                    continue

                # 开启 BasicAuth 鉴权
                if int(os.getenv("flyer_auth_enable", 0)) == 1:
                    cls.app.include_router(sub_module.router,
                                           prefix=f"{config.PREFIX}",
                                           dependencies=[Depends(authorize)])

                else:
                    cls.app.include_router(sub_module.router,
                                           prefix=f"{config.PREFIX}")

                logger.info(f"子项目加载成功：{dir.name}")

            except Exception:  # pylint: disable=broad-except
                logger.error(f"子路由加载错误：{traceback.format_exc()}")
                pass


@asynccontextmanager
async def lifespan(app: FastAPI):
    """生命周期管理
    """

    def on_startup():
        """启动时执行逻辑
        """
        max_threads = int(os.environ.get("flyer_threads", 5))
        logger.info(f"The Number of Threads per Worker: {max_threads}")
        loop = asyncio._get_running_loop()
        loop.set_default_executor(ThreadPoolExecutor(max_workers=max_threads))

        # 启动任务调度
        if not background_scheduler.running:
            logger.info("正在启动 Apscheduler 后台线程任务...")
            background_scheduler.start()

        # 启动异步任务调度
        if not asyncio_scheduler.running:
            logger.info("正在启动 Apscheduler 后台协程任务...")
            asyncio_scheduler.start()

    def on_shutdown():
        """关闭时执行逻辑
        """
        logger.warn("收到关闭信号，FastFlyer 开始执行优雅停止逻辑...")
        # 关闭任务调度器
        logger.info("正在关闭 Apscheduler 定时引擎...")
        background_scheduler.shutdown(wait=True)
        asyncio_scheduler.shutdown(wait=True)
        # 等待一段时间再退出（最小1秒，最大60秒）
        graceful_timeout = max(
            min(int(os.getenv("flyer_graceful_timeout", 1)), 60), 1)
        logger.warn(f"优雅停止逻辑执行完毕，FastFlyer 将在 {graceful_timeout}s 后自动退出...")
        for i in range(0, graceful_timeout):
            time.sleep(1)
            logger.warn(f"退出倒计时：{graceful_timeout - i}s ...")
        logger.warn("FastFlyer 已成功停止服务，感谢您的使用，再见！")

    try:
        on_startup()
        yield
    finally:
        on_shutdown()
